/*
  Copyright (c) 2007 Konosuke Watanabe <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h> 
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>

#include <libspe2.h>

#include "common.h"

#define MAX_SPES 6

#define DEFAULT_BENCHMARK_TIME 30

enum mode {
    MODE_DEFAULT,
    MODE_TRIP_VERIFY,
    MODE_BENCHMARK
};


static int mode = MODE_DEFAULT;

static unsigned char __A128(pattern[64][16]) = {{0}};
static unsigned int __A128(control_flag[4]) = {0};
static unsigned char __A128(initial_password[16]) = {0};
static result_t __A128(result[MAX_SPES][RESULT_BUF_LEN]);

/* tailˤͭʥǡμ / SPE  */
/* headˤͭʥǡƬ / PPE(1ɤ1䤹) */
/* ξԤƱ硤PPEɤ߽Фʤ */
static result_pointer_t __A128(result_tail[MAX_SPES]);
static result_pointer_t __A128(result_head[MAX_SPES]);
static long long int __A128(block_num_buf[2]) = {0};
static unsigned long long int __A128(processed_num_buf[MAX_SPES][2]) = {{0}};

typedef struct spe_thread_t {
    spe_context_ptr_t spe_ctx;
    pthread_t pthread;
} spe_thread_t;


static void *
spe_thread_function(void *arg)
{
    spe_thread_t *spe_thread = (spe_thread_t *)arg;
    spe_context_ptr_t context = spe_thread->spe_ctx;
    unsigned int entry = SPE_DEFAULT_ENTRY;
    spe_stop_info_t stop_info; 

    spe_context_run(context, &entry, 0, NULL, NULL, &stop_info);

    pthread_exit(NULL); 
} 


typedef struct checker_arg_t {
    int *flag;
    int mode;
    int num_spes;
    int quiet;
    FILE *out_file_fp;
    int ret;
} checker_arg_t;


#define PNUM_EVAL_INTERVAL 5

#define qprintf(...) {                          \
        if (! quiet) {                          \
            printf(__VA_ARGS__);                \
        }                                       \
    }

#define logprint(fp, ...) {                     \
        if (fp)                                 \
            fprintf(fp, __VA_ARGS__);           \
        qprintf(__VA_ARGS__);                   \
    }

#define delete_line() {                         \
        if (! quiet) {                          \
            printf("\x0d\x1b[K");               \
        }                                       \
    }


/* ̤ĥ */
void *
result_checker(void *arg)
{
    checker_arg_t checker_arg = *(checker_arg_t *)arg;
    int *check_flag = checker_arg.flag;
    int mode = checker_arg.mode;

    long long int match_num = 0;
    int num_spes = checker_arg.num_spes;
    int quiet = checker_arg.quiet;
    FILE *out = checker_arg.out_file_fp;

    int i;
    int loop = 2;

    struct timeval start, current, last;
    unsigned long long last_total_num = 0;
    unsigned long long total_num;
    char speed[128] = {0};
    int wait = 1;
    unsigned int interval;


    if (mode == MODE_DEFAULT) {
        logprint(out,
                 "#          trip    	password	salt	length	time\n");
        logprint(out, 
                 "------------------------------------------------------------"
                 "------------------\n");
    }


    gettimeofday(&start, NULL);
    last = start;

    while (loop) {
        int head[MAX_SPES], tail[MAX_SPES];
        result_t r;

        for (i = 0; i < num_spes; i ++) {
            /* ޤλǤκǿΥإåɤ */
            head[i] = result_head[i].p;
        }


        /* ٤ƤΥХåեåơ1ĤǤ⤢ȴ */
        while (wait) {            
            suseconds_t s;

            total_num = 0;
            for (i = 0; i < num_spes; i ++) {
                tail[i] = result_tail[i].p;
                total_num += processed_num_buf[i][0];
            }

            for (i = 0; i < num_spes; i ++) {
                if (head[i] != tail[i]) {
                    wait = 0;
                    break;
                }
            }


            if (mode == MODE_DEFAULT) {
                /*  */
                gettimeofday(&current, NULL);
                interval = current.tv_sec - last.tv_sec;
                s = current.tv_usec;
                if (s < last.tv_usec) {
                    interval --;
                    s += 1000 * 1000;
                }
                
                if (interval >= PNUM_EVAL_INTERVAL) {
                    sprintf(speed, "%dK trips/sec", 
                            (unsigned int)((float)(total_num - last_total_num) 
                                           / ((float)interval 
                                              + (float)(s - last.tv_usec) 
                                              / (1000.0 * 1000.0))) / 1000);

                    
                    delete_line();
                    qprintf("%s", speed);
                    
                    fflush(stdout);
                    last_total_num = total_num;
                    last = current;
                }
            }

            usleep(10);

            if (! *check_flag)
                break;

        }



        for (i = 0; i < num_spes; i ++) {
            while (head[i] != tail[i]) {
                
                r = result[i][head[i]];
                
                if (mode == MODE_DEFAULT || mode == MODE_TRIP_VERIFY) {

                    r.trip[11] = '\0';
                    r.password[8] = '\0';
                    r.salt[2] = '\0';


                    if (mode == MODE_TRIP_VERIFY) {
                        char *c, *host_trip;
                        c = crypt((char *)r.password, (char *)r.salt);
                        host_trip = c + 3;
                        if (strncmp(host_trip, (char *)r.trip,
                                    strlen((char *)r.trip)) != 0) {
                            logprint(out, 
                                     "FAILED:  password: %s (%d)   "
                                     "salt: %s   spu: %s   host: %s\n",
                                     r.password,
                                     (int)strlen((char *)r.password),
                                     r.salt, r.trip, host_trip);
                            checker_arg.ret = -1;
                        }

                    } else {
                        time_t t = time(NULL);
                        char time_buf[128];
                        strftime(time_buf, 256, "%Y-%m-%d %T", localtime(&t));

                        delete_line();

                        logprint(out, "%-10lld %s\t%-8s\t%s\t%d\t%s\n",
                                 match_num + 1, r.trip, r.password, r.salt,
                                 (int)strlen((char *)r.password), time_buf);

                        if (speed[0])
                            qprintf("%s", speed);
                    }

                    match_num ++;


                    fflush(stdout);
                }
                    
                head[i] ++;
                if (head[i] == RESULT_BUF_LEN)
                    head[i] = 0;
                    
                result_head[i].p = head[i];
            }
        }



        wait = 1;

        if (! *check_flag)
            loop --;

    }


    if (mode != MODE_TRIP_VERIFY) {
        float sec;
        /* ǽ */
        gettimeofday(&current, NULL);
        interval = current.tv_sec - start.tv_sec; 
        if (current.tv_usec < start.tv_usec) {
            interval --;
            current.tv_usec += 1000 * 1000;
        }

        sec = (float)interval 
            + (float)(current.tv_usec - start.tv_usec) / (1000.0 * 1000.0);
        delete_line();

        if (mode == MODE_DEFAULT) {
            printf("matched:  %lld trip(s)\n", match_num);
        }

        printf("searched: %lld trip(s)\n", total_num);
        printf("time: %f sec.\n", sec);

        printf("average: %dK trips/sec\n", 
               (unsigned int)((float)total_num / sec) / 1000);

        fflush(stdout);
    }
    
    pthread_exit(NULL);
}


int
setup_pattern(const char *lookup, int ignore_case)
{
    int i, j;
    unsigned char c;

    int len = strlen(lookup);
    unsigned char ptmp[10] = {0};

    if (len == 0) {
        printf("pattern is null\n");
        return -1;
    }

    if (len > 10) {
        printf("pattern is too long\n");
        return -1;
    }

    /* 1 ʸʸ */
    for (i = 0; i < 10; i ++) {
        c = lookup[i];

        if (c == '\0')
            break;

        /* ꤨʤʸäѲ */
        if (! (('.' <= c && c <= '9')
               || ('A' <= c && c <= 'Z')
               || ('a' <= c && c <= 'z'))) {
            printf("invalit character '%c' in pattern\n", c);
            return -1;
        }

        if (c >= 'a') {
            if (ignore_case) {
                /* ʸʸ̤ʤ a  A ˤ */
                c -= ('a' - 'A');
            } else {
                /* ʸʸ̤ */
                c -= ('a' - ('Z' + 1));
            }
        }

        if (c > '9')
            c -= ('A' - ('9' + 1));

        c -= '.';

        /* Ǹ1ʸβ2bit̵뤵Τǻ꤬Фʤ */
        if (i == 9 && (c & 3) != 0) {
            printf("last character ('%c') is invalid\n", lookup[i]);
            return -1;
        }

        ptmp[i] = c;
    }



    for (i = 0; i < 10; i ++) {
        for (j = 0; j < 6; j ++) {
            if ((ptmp[i] >> (5 - j)) & 1) {
                memset(pattern[i * 6 + j], 0xff, 16);
                /* printf("%d: 11\n", i * 6 + j); */
            } else {
                memset(pattern[i * 6 + j], 0, 16);
                /* printf("%d: 00\n", i * 6 + j); */
            }
        }
    }

    return len;
}


static int
mbox_write(spe_context_ptr_t spe_ctx, unsigned int data)
{
    int ret;
    while((ret = spe_in_mbox_status(spe_ctx)) < 1) {
        if (ret < 0)
            return ret;
    }

    return spe_in_mbox_write(spe_ctx, &data, 1, SPE_MBOX_ANY_NONBLOCKING);
}

static void
gen_random_password(unsigned char *password)
{
    int i;

    for (i = 0; i < 5; i ++)
        password[i] = (rand() >> 16) % (unsigned char)('~' - ' ' + 1) + ' ';
    password[5] = ' ';
    password[6] = ' ';
    password[7] = ' ';
}

static int
increment_password(unsigned char *password, int digit)
{

    password[digit] ++;

    if (password[digit] > '~') {
        password[digit] = ' ';
        if (digit > 0)
            return increment_password(password, digit - 1);
        else
            memset(password, ' ', 8);
    }
    
    return digit;
}

static void
usage(char *cmd)
{
    printf("Usage: %s [OPTION] STRING\n"
           "Search 2ch trip.\n\n"

           "  -n SPENUM      number of using SPEs\n"
           "  -p PASSWORD    set the starting point of the search\n"
           "  -B BLOCKNUM    limit search range\n"
           "  -i             ignore case\n"
           "  -k             kana mode (always use '..' as a salt)\n"
           "  -b             run benchmark\n"
           "  -T SEC         set benchmark length\n"
           "  -o FILENAME    output to file\n"
           "  -q             quiet mode\n"
           "  -t             test mode (for debug)\n"
           "  -v             show version\n"
           "  -h             show this message\n\n"
           , cmd);
}


static void
version()
{
    printf("celltripper %s\n", VERSION);
}


int
main(int argc, char *argv[])
{ 
    int opt;
    spe_program_handle_t *program;

    pthread_t result_check_thread;;

    spe_thread_t spe_thread[MAX_SPES];
    int i;

    int flags = 0; 

    int ignore_case = 0;

    int pattern_len = 0;
    int set_password = 0;

    int kana_mode = 0;

    int check_flag = 1;
    checker_arg_t checker_arg;

    int num_spes = 1;
    long long int block_num = -1;
    int quiet = 0;

    int benchmark_time = DEFAULT_BENCHMARK_TIME;

    int out_file = 0;
    char out_file_name[256] = {0};
    FILE *out_file_fp = NULL;

    srand(time(NULL) ^ getpid());

    gen_random_password(initial_password);

    while ((opt = getopt(argc, argv, "p:kB:n:ibT:o:qtvh")) != -1) {
        switch (opt) {
        case 'p':
            /* initial password */
            strncpy((void *)initial_password, (void *)optarg, 5);

            for (i = strlen((char *)initial_password); i < 8; i ++) {
                initial_password[i] = ' ';
            }

            initial_password[8] = '\0';
            set_password = 1;
            break;
        case 'k':
            /* kana mode */
            kana_mode = 1;
            break;
        case 'n':
            num_spes = atoi(optarg);
            break;
        case 'i':
            ignore_case = 1;
            break;
        case 'b':
            mode = MODE_BENCHMARK;
            strcpy((char *)initial_password, "Cell    ");
            pattern_len = setup_pattern("Cell", 0);
            break;
        case 't':
            mode = MODE_TRIP_VERIFY;
            break;
        case 'B':
            block_num = atoll(optarg);
            break;
        case 'T':
            benchmark_time = atoi(optarg);
            break;
        case 'o':
            out_file = 1;
            strncpy(out_file_name, (void *)optarg, 
                    sizeof(out_file_name) - 1);
            break;
        case 'q':
            quiet = 1;
            break;
        case 'v':
            /* show version */
            version();
            return 1;
        case 'h':
        default:
            /* help */
            usage(argv[0]);
            return 1;
        }
    }


    if (num_spes <= 0 || num_spes > MAX_SPES) {
        printf("invalid number of SPEs (%d)\n", num_spes);
        return 1;
    }


    if (argv[optind] == NULL) {
        /* Ǹ˰ʤä饨顼 */
        if (mode == MODE_DEFAULT) {
            usage(argv[0]);
            return 1;
        }
    } else {
        if (mode != MODE_TRIP_VERIFY) {
            /* pattern */
            if ((pattern_len = setup_pattern(argv[optind], ignore_case)) < 0) {
                printf("bad pattern\n");
                return 1;
            }
        }
    }


    if (out_file && (out_file_fp = fopen(out_file_name, "w")) == NULL) {
        perror("open output file");
        return 1;
    }


    program = spe_image_open(SPE_IMAGE);

    qprintf("number of SPE(s): %d\n", num_spes);

    qprintf("start SPE thread(s)\n");

    for (i = 0; i < num_spes; i ++) {

        spe_thread[i].spe_ctx = spe_context_create(flags, NULL); 

        spe_program_load(spe_thread[i].spe_ctx, program); 

        pthread_create(&spe_thread[i].pthread, NULL, &spe_thread_function, 
                       &spe_thread[i]); 

    }

    qprintf("initialize SPE thread(s)\n");


    block_num_buf[0] = block_num;

    for (i = 0; i < num_spes; i ++) {
        result_tail[i].p = 0;
        result_head[i].p = 0;


        mbox_write(spe_thread[i].spe_ctx, CMD_SET_ID);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)i);


        mbox_write(spe_thread[i].spe_ctx, CMD_SET_CONTROL_EA);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)&control_flag);

        mbox_write(spe_thread[i].spe_ctx, CMD_SET_RESULT_EA);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)result[i]);

        mbox_write(spe_thread[i].spe_ctx, CMD_SET_HEAD_PTR_EA);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)&result_head[i]);

        mbox_write(spe_thread[i].spe_ctx, CMD_SET_TAIL_PTR_EA);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)&result_tail[i]);

        mbox_write(spe_thread[i].spe_ctx, CMD_SET_PROC_NUM_EA);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)processed_num_buf[i]);

        if (ignore_case)
            mbox_write(spe_thread[i].spe_ctx, CMD_SET_CASE_FLAG);

        if (kana_mode)
            mbox_write(spe_thread[i].spe_ctx, CMD_SET_KANA_MODE);

        /* SPEʬδֳ֤ǽ */
        mbox_write(spe_thread[i].spe_ctx, CMD_SET_DISTANCE);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)num_spes);

        mbox_write(spe_thread[i].spe_ctx, CMD_LOAD_PATTERN);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)pattern);

        mbox_write(spe_thread[i].spe_ctx, CMD_SET_PATTERN_LEN);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)pattern_len);

        mbox_write(spe_thread[i].spe_ctx, CMD_LOAD_BLOCK_NUM);
        mbox_write(spe_thread[i].spe_ctx, (unsigned int)block_num_buf);
    }


    checker_arg.flag = &check_flag;
    checker_arg.mode = mode;
    checker_arg.num_spes = num_spes;
    checker_arg.out_file_fp = out_file_fp;
    checker_arg.quiet = quiet;
    checker_arg.ret = 0;

    /* Ͻå */
    pthread_create(&result_check_thread, NULL, &result_checker, &checker_arg); 

    if (mode == MODE_TRIP_VERIFY) {
        /* ⡼ */

        /* ϥѥ */
        qprintf("start test (password: \"%s\")\n", initial_password);

        for (i = 0; i < num_spes; i ++) {
            mbox_write(spe_thread[i].spe_ctx, CMD_LOAD_INIT_PASS);
            mbox_write(spe_thread[i].spe_ctx, (unsigned int)initial_password);

            /* ȥåץƥȥ⡼ON */
            mbox_write(spe_thread[i].spe_ctx, CMD_SET_PATTERN_LEN);
            mbox_write(spe_thread[i].spe_ctx, (unsigned int)0);

            /* ư */
            mbox_write(spe_thread[i].spe_ctx, CMD_RUN);

            if (increment_password(initial_password, 5) != 5)
                break;
        }

        /* λؼ */
        for (i = 0; i < num_spes; i ++)
            mbox_write(spe_thread[i].spe_ctx, CMD_EXIT);

    } else if (mode == MODE_BENCHMARK) {
        /* ٥ޡ⡼ */

        struct timespec t;

        qprintf("start benchmark\n");

        /* ϥѥ */
        for (i = 0; i < num_spes; i ++) {
            mbox_write(spe_thread[i].spe_ctx, CMD_LOAD_INIT_PASS);
            mbox_write(spe_thread[i].spe_ctx, (unsigned int)initial_password);

            mbox_write(spe_thread[i].spe_ctx, CMD_RUN);

            increment_password(initial_password, 5);
        }

        t.tv_sec = benchmark_time;
        t.tv_nsec = 0;

        qprintf("please wait %d seconds\n", benchmark_time);
        nanosleep(&t, NULL);

        control_flag[0] = CONTROL_STOP;

        for (i = 0; i < num_spes; i ++)
            mbox_write(spe_thread[i].spe_ctx, CMD_EXIT);

    } else {
        /* ̾⡼ */

        qprintf("start normal search\n");
           
        /* ϥѥ */
        for (i = 0; i < num_spes; i ++) {
            mbox_write(spe_thread[i].spe_ctx, CMD_LOAD_INIT_PASS);
            mbox_write(spe_thread[i].spe_ctx, (unsigned int)initial_password);

            mbox_write(spe_thread[i].spe_ctx, CMD_RUN);
            
            if (increment_password(initial_password, 5) != 5)
                if (-- block_num == 0) 
                    break;
        }

        if (block_num >= 0)
            for (i = 0; i < num_spes; i ++)
                mbox_write(spe_thread[i].spe_ctx, CMD_EXIT);
    }


    for (i = 0; i < num_spes; i ++) {
        fflush(stdout);
        pthread_join(spe_thread[i].pthread, NULL); 
        spe_context_destroy(spe_thread[i].spe_ctx); 
    }

    qprintf("wait checker thread\n");
    check_flag = 0;
    pthread_join(result_check_thread, NULL); 



    if (mode == MODE_TRIP_VERIFY) {
        if (checker_arg.ret < 0) {
            printf("test failed\n");
            return 1;
        } else {
            printf("all trips are OK\n");
        }
    }

    qprintf("done\n");

    return 0;
}
