/*
  Copyright (c) 2007 Konosuke Watanabe <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

static char
fix_invalid_salt(char c)
{
    if (! (isalnum(c) || c == '.' || c == '/'))
        return '.';

    return c;
}


static void
check_trip(char *password, char *salt, 
           char *pattern, int ignore_case, int test_mode)
{
    char *match = NULL;
    char *c, *trip;

    c = crypt(password, salt);
    trip = c + 3;

    if (test_mode)
        match = c;
    else if (ignore_case)
        match = strcasestr(trip, pattern);
    else 
        match = strstr(trip, pattern);

    if (match) {
        printf("%s\t%s\t%s\t%d\n",
               trip, password, salt, strlen(password));
    }

}

static int
increment_password(char *password, int digit)
{

    password[digit] ++;

    if (password[digit] > '~') {
        password[digit] = ' ';
        if (digit > 0)
            return increment_password(password, digit - 1);
        else
            memset(password, ' ', 8);
    }
    
    return digit;
}

int
main(int argc, char *argv[])
{
    int i, j, k, m;
    int opt;

    int block_num = 1;
    int test_mode = 0;
    int kana_mode = 0;
    
    char password[9] = "celltrip";
    char pattern[11];
    char salt[3] = {0};
    int ignore_case = 0;

    while ((opt = getopt(argc, argv, "p:B:tik")) != -1) {
        switch (opt) {
        case 'p':
            /* initial password */
            strncpy((void *)password, (void *)optarg, 8);
            for (i = strlen((char *)password); i < 8; i ++) {
                password[i] = ' ';
            }
            password[8] = '\0';
            break;
        case 'i':
            ignore_case = 1;
            break;
        case 'B':
            block_num = atoi(optarg);
            break;
        case 't':
            // show all
            test_mode = 1;
            break;
        case 'k':
            kana_mode = 1;
            break;
        default:
            return 1;
        }
    }

    if (! test_mode) {
        if (argv[optind] == NULL) {
            printf("no pattern\n");
            return 1;
        }

        strncpy(pattern, argv[optind], 11);

        if (strlen(pattern) == 0) {
            printf("invalid pattern\n");
            return 1;
        }
    }

    salt[0] = kana_mode ? '.' : fix_invalid_salt(password[1]);
    salt[1] = kana_mode ? '.' : fix_invalid_salt(password[2]);

    password[8] = (char)'\0';

    for (i = 0; i < block_num; i ++) {
        for (j = ' '; j <= '~'; j ++) {
            password[5] = (char)j;
            for (k = ' '; k <= '~'; k ++) {
                password[6] = (char)k;
                for (m = ' '; m <= '~'; m ++) {
                    password[7] = (char)m;
                    check_trip(password, salt, pattern, ignore_case, test_mode);
                }
                password[7] = (char)'\0';
                check_trip(password, salt, pattern, ignore_case, test_mode);
            }
        }
        increment_password(password, 4);
    }
    
    return 0;
}
