/*
  Copyright (c) 2007 Konosuke Watanabe <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <spu_mfcio.h>
#include <spu_intrinsics.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "spe.h"

#include "spu_nonstd.c"

static vec_uchar16 __A128(pattern[64]) = {{0}};
static unsigned int __A128(control_flag[4]) = {0};
static unsigned char __A128(ipassword[16]) = {0};

static result_pointer_t __A128(tail_pointer);
static result_pointer_t __A128(head_pointer);
static result_t __A128(result);
static long long int __A128(block_num_buf[2]);
static unsigned long long __A128(proc_num_buf[2]) = {0};

typedef struct host_ea {
    unsigned int tail_pointer_ea;
    unsigned int head_pointer_ea;
    unsigned int result_buf_ea;
    unsigned int proc_num_ea;
} host_ea;

static int id = -1;

//#define SPE_DEBUG

#ifdef DEBUG
static void
dump_vector(unsigned char *head, vec_uchar16 vec, int num)
{
    int i;
    unsigned char c;

    for (i = 0; i < num && i < 128; i ++) {
        c = (spu_extract(vec, i / 8) >> (i % 8)) & 1;

        if (i % 64 == 0)
            printf("%s: ", head);

        printf("%d", c);
        if (i == 63 && num > 64)
            printf("\n");
        else if (i % 8 == 7)
            printf(" ");
    }
    printf("\n");
}
#endif


static void
trip_search_core(const vec_uchar16 *passwd, const vec_uchar16 *salt_flag,
                 const vec_uchar16 *pattern, const int pattern_len,
                 const int ignore_case, const vec_uchar16 valid_range, 
                 const unsigned char *password, const int offset, 
                 const int *salt, const host_ea ea)
{
    int i, j, k;
    int m;

    vec_uchar16 ks[16][48];

    vec_uchar16 block[66];

    vec_uchar16 *R0, *R1, *vptmp;

    vec_uchar16 r0[32] = {{0}};
    vec_uchar16 r1[32] = {{0}};

    unsigned int tail_pointer_ea = ea.tail_pointer_ea;
    unsigned int head_pointer_ea = ea.head_pointer_ea;
    unsigned int result_buf_ea = ea.result_buf_ea;

    R0 = r0;
    R1 = r1;

    block[64] = VZERO;
    block[65] = VZERO;

    ks[0][0] = passwd[13];
    ks[0][1] = passwd[52];
    ks[0][2] = passwd[37];
    ks[0][3] = passwd[59];
    ks[0][4] = passwd[54];
    ks[0][5] = passwd[22];
    ks[0][6] = passwd[38];
    ks[0][7] = passwd[62];
    ks[0][8] = passwd[5];
    ks[0][9] = passwd[14];
    ks[0][10] = passwd[20];
    ks[0][11] = passwd[45];
    ks[0][12] = passwd[4];
    ks[0][13] = passwd[36];
    ks[0][14] = passwd[29];
    ks[0][15] = passwd[30];
    ks[0][16] = passwd[43];
    ks[0][17] = passwd[61];
    ks[0][18] = passwd[60];
    ks[0][19] = passwd[6];
    ks[0][20] = passwd[35];
    ks[0][21] = passwd[28];
    ks[0][22] = passwd[21];
    ks[0][23] = passwd[46];
    ks[0][24] = passwd[17];
    ks[0][25] = passwd[27];
    ks[0][26] = passwd[32];
    ks[0][27] = passwd[49];
    ks[0][28] = passwd[34];
    ks[0][29] = passwd[3];
    ks[0][30] = passwd[40];
    ks[0][31] = passwd[25];
    ks[0][32] = passwd[2];
    ks[0][33] = passwd[50];
    ks[0][34] = passwd[16];
    ks[0][35] = passwd[26];
    ks[0][36] = passwd[58];
    ks[0][37] = passwd[18];
    ks[0][38] = passwd[33];
    ks[0][39] = passwd[56];
    ks[0][40] = passwd[8];
    ks[0][41] = passwd[19];
    ks[0][42] = passwd[42];
    ks[0][43] = passwd[9];
    ks[0][44] = passwd[10];
    ks[0][45] = passwd[57];
    ks[0][46] = passwd[48];
    ks[0][47] = passwd[24];
    ks[1][0] = passwd[5];
    ks[1][1] = passwd[44];
    ks[1][2] = passwd[29];
    ks[1][3] = passwd[51];
    ks[1][4] = passwd[46];
    ks[1][5] = passwd[14];
    ks[1][6] = passwd[30];
    ks[1][7] = passwd[54];
    ks[1][8] = passwd[60];
    ks[1][9] = passwd[6];
    ks[1][10] = passwd[12];
    ks[1][11] = passwd[37];
    ks[1][12] = passwd[59];
    ks[1][13] = passwd[28];
    ks[1][14] = passwd[21];
    ks[1][15] = passwd[22];
    ks[1][16] = passwd[35];
    ks[1][17] = passwd[53];
    ks[1][18] = passwd[52];
    ks[1][19] = passwd[61];
    ks[1][20] = passwd[62];
    ks[1][21] = passwd[20];
    ks[1][22] = passwd[13];
    ks[1][23] = passwd[38];
    ks[1][24] = passwd[9];
    ks[1][25] = passwd[19];
    ks[1][26] = passwd[24];
    ks[1][27] = passwd[41];
    ks[1][28] = passwd[26];
    ks[1][29] = passwd[56];
    ks[1][30] = passwd[32];
    ks[1][31] = passwd[17];
    ks[1][32] = passwd[27];
    ks[1][33] = passwd[42];
    ks[1][34] = passwd[8];
    ks[1][35] = passwd[18];
    ks[1][36] = passwd[50];
    ks[1][37] = passwd[10];
    ks[1][38] = passwd[25];
    ks[1][39] = passwd[48];
    ks[1][40] = passwd[0];
    ks[1][41] = passwd[11];
    ks[1][42] = passwd[34];
    ks[1][43] = passwd[1];
    ks[1][44] = passwd[2];
    ks[1][45] = passwd[49];
    ks[1][46] = passwd[40];
    ks[1][47] = passwd[16];
    ks[2][0] = passwd[52];
    ks[2][1] = passwd[28];
    ks[2][2] = passwd[13];
    ks[2][3] = passwd[35];
    ks[2][4] = passwd[30];
    ks[2][5] = passwd[61];
    ks[2][6] = passwd[14];
    ks[2][7] = passwd[38];
    ks[2][8] = passwd[44];
    ks[2][9] = passwd[53];
    ks[2][10] = passwd[59];
    ks[2][11] = passwd[21];
    ks[2][12] = passwd[43];
    ks[2][13] = passwd[12];
    ks[2][14] = passwd[5];
    ks[2][15] = passwd[6];
    ks[2][16] = passwd[54];
    ks[2][17] = passwd[37];
    ks[2][18] = passwd[36];
    ks[2][19] = passwd[45];
    ks[2][20] = passwd[46];
    ks[2][21] = passwd[4];
    ks[2][22] = passwd[60];
    ks[2][23] = passwd[22];
    ks[2][24] = passwd[58];
    ks[2][25] = passwd[3];
    ks[2][26] = passwd[8];
    ks[2][27] = passwd[25];
    ks[2][28] = passwd[10];
    ks[2][29] = passwd[40];
    ks[2][30] = passwd[16];
    ks[2][31] = passwd[1];
    ks[2][32] = passwd[11];
    ks[2][33] = passwd[26];
    ks[2][34] = passwd[57];
    ks[2][35] = passwd[2];
    ks[2][36] = passwd[34];
    ks[2][37] = passwd[27];
    ks[2][38] = passwd[9];
    ks[2][39] = passwd[32];
    ks[2][40] = passwd[49];
    ks[2][41] = passwd[56];
    ks[2][42] = passwd[18];
    ks[2][43] = passwd[50];
    ks[2][44] = passwd[19];
    ks[2][45] = passwd[33];
    ks[2][46] = passwd[24];
    ks[2][47] = passwd[0];
    ks[3][0] = passwd[36];
    ks[3][1] = passwd[12];
    ks[3][2] = passwd[60];
    ks[3][3] = passwd[54];
    ks[3][4] = passwd[14];
    ks[3][5] = passwd[45];
    ks[3][6] = passwd[61];
    ks[3][7] = passwd[22];
    ks[3][8] = passwd[28];
    ks[3][9] = passwd[37];
    ks[3][10] = passwd[43];
    ks[3][11] = passwd[5];
    ks[3][12] = passwd[62];
    ks[3][13] = passwd[59];
    ks[3][14] = passwd[52];
    ks[3][15] = passwd[53];
    ks[3][16] = passwd[38];
    ks[3][17] = passwd[21];
    ks[3][18] = passwd[20];
    ks[3][19] = passwd[29];
    ks[3][20] = passwd[30];
    ks[3][21] = passwd[51];
    ks[3][22] = passwd[44];
    ks[3][23] = passwd[6];
    ks[3][24] = passwd[42];
    ks[3][25] = passwd[48];
    ks[3][26] = passwd[57];
    ks[3][27] = passwd[9];
    ks[3][28] = passwd[27];
    ks[3][29] = passwd[24];
    ks[3][30] = passwd[0];
    ks[3][31] = passwd[50];
    ks[3][32] = passwd[56];
    ks[3][33] = passwd[10];
    ks[3][34] = passwd[41];
    ks[3][35] = passwd[19];
    ks[3][36] = passwd[18];
    ks[3][37] = passwd[11];
    ks[3][38] = passwd[58];
    ks[3][39] = passwd[16];
    ks[3][40] = passwd[33];
    ks[3][41] = passwd[40];
    ks[3][42] = passwd[2];
    ks[3][43] = passwd[34];
    ks[3][44] = passwd[3];
    ks[3][45] = passwd[17];
    ks[3][46] = passwd[8];
    ks[3][47] = passwd[49];
    ks[4][0] = passwd[20];
    ks[4][1] = passwd[59];
    ks[4][2] = passwd[44];
    ks[4][3] = passwd[38];
    ks[4][4] = passwd[61];
    ks[4][5] = passwd[29];
    ks[4][6] = passwd[45];
    ks[4][7] = passwd[6];
    ks[4][8] = passwd[12];
    ks[4][9] = passwd[21];
    ks[4][10] = passwd[62];
    ks[4][11] = passwd[52];
    ks[4][12] = passwd[46];
    ks[4][13] = passwd[43];
    ks[4][14] = passwd[36];
    ks[4][15] = passwd[37];
    ks[4][16] = passwd[22];
    ks[4][17] = passwd[5];
    ks[4][18] = passwd[4];
    ks[4][19] = passwd[13];
    ks[4][20] = passwd[14];
    ks[4][21] = passwd[35];
    ks[4][22] = passwd[28];
    ks[4][23] = passwd[53];
    ks[4][24] = passwd[26];
    ks[4][25] = passwd[32];
    ks[4][26] = passwd[41];
    ks[4][27] = passwd[58];
    ks[4][28] = passwd[11];
    ks[4][29] = passwd[8];
    ks[4][30] = passwd[49];
    ks[4][31] = passwd[34];
    ks[4][32] = passwd[40];
    ks[4][33] = passwd[27];
    ks[4][34] = passwd[25];
    ks[4][35] = passwd[3];
    ks[4][36] = passwd[2];
    ks[4][37] = passwd[56];
    ks[4][38] = passwd[42];
    ks[4][39] = passwd[0];
    ks[4][40] = passwd[17];
    ks[4][41] = passwd[24];
    ks[4][42] = passwd[19];
    ks[4][43] = passwd[18];
    ks[4][44] = passwd[48];
    ks[4][45] = passwd[1];
    ks[4][46] = passwd[57];
    ks[4][47] = passwd[33];
    ks[5][0] = passwd[4];
    ks[5][1] = passwd[43];
    ks[5][2] = passwd[28];
    ks[5][3] = passwd[22];
    ks[5][4] = passwd[45];
    ks[5][5] = passwd[13];
    ks[5][6] = passwd[29];
    ks[5][7] = passwd[53];
    ks[5][8] = passwd[59];
    ks[5][9] = passwd[5];
    ks[5][10] = passwd[46];
    ks[5][11] = passwd[36];
    ks[5][12] = passwd[30];
    ks[5][13] = passwd[62];
    ks[5][14] = passwd[20];
    ks[5][15] = passwd[21];
    ks[5][16] = passwd[6];
    ks[5][17] = passwd[52];
    ks[5][18] = passwd[51];
    ks[5][19] = passwd[60];
    ks[5][20] = passwd[61];
    ks[5][21] = passwd[54];
    ks[5][22] = passwd[12];
    ks[5][23] = passwd[37];
    ks[5][24] = passwd[10];
    ks[5][25] = passwd[16];
    ks[5][26] = passwd[25];
    ks[5][27] = passwd[42];
    ks[5][28] = passwd[56];
    ks[5][29] = passwd[57];
    ks[5][30] = passwd[33];
    ks[5][31] = passwd[18];
    ks[5][32] = passwd[24];
    ks[5][33] = passwd[11];
    ks[5][34] = passwd[9];
    ks[5][35] = passwd[48];
    ks[5][36] = passwd[19];
    ks[5][37] = passwd[40];
    ks[5][38] = passwd[26];
    ks[5][39] = passwd[49];
    ks[5][40] = passwd[1];
    ks[5][41] = passwd[8];
    ks[5][42] = passwd[3];
    ks[5][43] = passwd[2];
    ks[5][44] = passwd[32];
    ks[5][45] = passwd[50];
    ks[5][46] = passwd[41];
    ks[5][47] = passwd[17];
    ks[6][0] = passwd[51];
    ks[6][1] = passwd[62];
    ks[6][2] = passwd[12];
    ks[6][3] = passwd[6];
    ks[6][4] = passwd[29];
    ks[6][5] = passwd[60];
    ks[6][6] = passwd[13];
    ks[6][7] = passwd[37];
    ks[6][8] = passwd[43];
    ks[6][9] = passwd[52];
    ks[6][10] = passwd[30];
    ks[6][11] = passwd[20];
    ks[6][12] = passwd[14];
    ks[6][13] = passwd[46];
    ks[6][14] = passwd[4];
    ks[6][15] = passwd[5];
    ks[6][16] = passwd[53];
    ks[6][17] = passwd[36];
    ks[6][18] = passwd[35];
    ks[6][19] = passwd[44];
    ks[6][20] = passwd[45];
    ks[6][21] = passwd[38];
    ks[6][22] = passwd[59];
    ks[6][23] = passwd[21];
    ks[6][24] = passwd[27];
    ks[6][25] = passwd[0];
    ks[6][26] = passwd[9];
    ks[6][27] = passwd[26];
    ks[6][28] = passwd[40];
    ks[6][29] = passwd[41];
    ks[6][30] = passwd[17];
    ks[6][31] = passwd[2];
    ks[6][32] = passwd[8];
    ks[6][33] = passwd[56];
    ks[6][34] = passwd[58];
    ks[6][35] = passwd[32];
    ks[6][36] = passwd[3];
    ks[6][37] = passwd[24];
    ks[6][38] = passwd[10];
    ks[6][39] = passwd[33];
    ks[6][40] = passwd[50];
    ks[6][41] = passwd[57];
    ks[6][42] = passwd[48];
    ks[6][43] = passwd[19];
    ks[6][44] = passwd[16];
    ks[6][45] = passwd[34];
    ks[6][46] = passwd[25];
    ks[6][47] = passwd[1];
    ks[7][0] = passwd[35];
    ks[7][1] = passwd[46];
    ks[7][2] = passwd[59];
    ks[7][3] = passwd[53];
    ks[7][4] = passwd[13];
    ks[7][5] = passwd[44];
    ks[7][6] = passwd[60];
    ks[7][7] = passwd[21];
    ks[7][8] = passwd[62];
    ks[7][9] = passwd[36];
    ks[7][10] = passwd[14];
    ks[7][11] = passwd[4];
    ks[7][12] = passwd[61];
    ks[7][13] = passwd[30];
    ks[7][14] = passwd[51];
    ks[7][15] = passwd[52];
    ks[7][16] = passwd[37];
    ks[7][17] = passwd[20];
    ks[7][18] = passwd[54];
    ks[7][19] = passwd[28];
    ks[7][20] = passwd[29];
    ks[7][21] = passwd[22];
    ks[7][22] = passwd[43];
    ks[7][23] = passwd[5];
    ks[7][24] = passwd[11];
    ks[7][25] = passwd[49];
    ks[7][26] = passwd[58];
    ks[7][27] = passwd[10];
    ks[7][28] = passwd[24];
    ks[7][29] = passwd[25];
    ks[7][30] = passwd[1];
    ks[7][31] = passwd[19];
    ks[7][32] = passwd[57];
    ks[7][33] = passwd[40];
    ks[7][34] = passwd[42];
    ks[7][35] = passwd[16];
    ks[7][36] = passwd[48];
    ks[7][37] = passwd[8];
    ks[7][38] = passwd[27];
    ks[7][39] = passwd[17];
    ks[7][40] = passwd[34];
    ks[7][41] = passwd[41];
    ks[7][42] = passwd[32];
    ks[7][43] = passwd[3];
    ks[7][44] = passwd[0];
    ks[7][45] = passwd[18];
    ks[7][46] = passwd[9];
    ks[7][47] = passwd[50];
    ks[8][0] = passwd[62];
    ks[8][1] = passwd[38];
    ks[8][2] = passwd[51];
    ks[8][3] = passwd[45];
    ks[8][4] = passwd[5];
    ks[8][5] = passwd[36];
    ks[8][6] = passwd[52];
    ks[8][7] = passwd[13];
    ks[8][8] = passwd[54];
    ks[8][9] = passwd[28];
    ks[8][10] = passwd[6];
    ks[8][11] = passwd[59];
    ks[8][12] = passwd[53];
    ks[8][13] = passwd[22];
    ks[8][14] = passwd[43];
    ks[8][15] = passwd[44];
    ks[8][16] = passwd[29];
    ks[8][17] = passwd[12];
    ks[8][18] = passwd[46];
    ks[8][19] = passwd[20];
    ks[8][20] = passwd[21];
    ks[8][21] = passwd[14];
    ks[8][22] = passwd[35];
    ks[8][23] = passwd[60];
    ks[8][24] = passwd[3];
    ks[8][25] = passwd[41];
    ks[8][26] = passwd[50];
    ks[8][27] = passwd[2];
    ks[8][28] = passwd[16];
    ks[8][29] = passwd[17];
    ks[8][30] = passwd[58];
    ks[8][31] = passwd[11];
    ks[8][32] = passwd[49];
    ks[8][33] = passwd[32];
    ks[8][34] = passwd[34];
    ks[8][35] = passwd[8];
    ks[8][36] = passwd[40];
    ks[8][37] = passwd[0];
    ks[8][38] = passwd[19];
    ks[8][39] = passwd[9];
    ks[8][40] = passwd[26];
    ks[8][41] = passwd[33];
    ks[8][42] = passwd[24];
    ks[8][43] = passwd[56];
    ks[8][44] = passwd[57];
    ks[8][45] = passwd[10];
    ks[8][46] = passwd[1];
    ks[8][47] = passwd[42];
    ks[9][0] = passwd[46];
    ks[9][1] = passwd[22];
    ks[9][2] = passwd[35];
    ks[9][3] = passwd[29];
    ks[9][4] = passwd[52];
    ks[9][5] = passwd[20];
    ks[9][6] = passwd[36];
    ks[9][7] = passwd[60];
    ks[9][8] = passwd[38];
    ks[9][9] = passwd[12];
    ks[9][10] = passwd[53];
    ks[9][11] = passwd[43];
    ks[9][12] = passwd[37];
    ks[9][13] = passwd[6];
    ks[9][14] = passwd[62];
    ks[9][15] = passwd[28];
    ks[9][16] = passwd[13];
    ks[9][17] = passwd[59];
    ks[9][18] = passwd[30];
    ks[9][19] = passwd[4];
    ks[9][20] = passwd[5];
    ks[9][21] = passwd[61];
    ks[9][22] = passwd[54];
    ks[9][23] = passwd[44];
    ks[9][24] = passwd[48];
    ks[9][25] = passwd[25];
    ks[9][26] = passwd[34];
    ks[9][27] = passwd[19];
    ks[9][28] = passwd[0];
    ks[9][29] = passwd[1];
    ks[9][30] = passwd[42];
    ks[9][31] = passwd[56];
    ks[9][32] = passwd[33];
    ks[9][33] = passwd[16];
    ks[9][34] = passwd[18];
    ks[9][35] = passwd[57];
    ks[9][36] = passwd[24];
    ks[9][37] = passwd[49];
    ks[9][38] = passwd[3];
    ks[9][39] = passwd[58];
    ks[9][40] = passwd[10];
    ks[9][41] = passwd[17];
    ks[9][42] = passwd[8];
    ks[9][43] = passwd[40];
    ks[9][44] = passwd[41];
    ks[9][45] = passwd[27];
    ks[9][46] = passwd[50];
    ks[9][47] = passwd[26];
    ks[10][0] = passwd[30];
    ks[10][1] = passwd[6];
    ks[10][2] = passwd[54];
    ks[10][3] = passwd[13];
    ks[10][4] = passwd[36];
    ks[10][5] = passwd[4];
    ks[10][6] = passwd[20];
    ks[10][7] = passwd[44];
    ks[10][8] = passwd[22];
    ks[10][9] = passwd[59];
    ks[10][10] = passwd[37];
    ks[10][11] = passwd[62];
    ks[10][12] = passwd[21];
    ks[10][13] = passwd[53];
    ks[10][14] = passwd[46];
    ks[10][15] = passwd[12];
    ks[10][16] = passwd[60];
    ks[10][17] = passwd[43];
    ks[10][18] = passwd[14];
    ks[10][19] = passwd[51];
    ks[10][20] = passwd[52];
    ks[10][21] = passwd[45];
    ks[10][22] = passwd[38];
    ks[10][23] = passwd[28];
    ks[10][24] = passwd[32];
    ks[10][25] = passwd[9];
    ks[10][26] = passwd[18];
    ks[10][27] = passwd[3];
    ks[10][28] = passwd[49];
    ks[10][29] = passwd[50];
    ks[10][30] = passwd[26];
    ks[10][31] = passwd[40];
    ks[10][32] = passwd[17];
    ks[10][33] = passwd[0];
    ks[10][34] = passwd[2];
    ks[10][35] = passwd[41];
    ks[10][36] = passwd[8];
    ks[10][37] = passwd[33];
    ks[10][38] = passwd[48];
    ks[10][39] = passwd[42];
    ks[10][40] = passwd[27];
    ks[10][41] = passwd[1];
    ks[10][42] = passwd[57];
    ks[10][43] = passwd[24];
    ks[10][44] = passwd[25];
    ks[10][45] = passwd[11];
    ks[10][46] = passwd[34];
    ks[10][47] = passwd[10];
    ks[11][0] = passwd[14];
    ks[11][1] = passwd[53];
    ks[11][2] = passwd[38];
    ks[11][3] = passwd[60];
    ks[11][4] = passwd[20];
    ks[11][5] = passwd[51];
    ks[11][6] = passwd[4];
    ks[11][7] = passwd[28];
    ks[11][8] = passwd[6];
    ks[11][9] = passwd[43];
    ks[11][10] = passwd[21];
    ks[11][11] = passwd[46];
    ks[11][12] = passwd[5];
    ks[11][13] = passwd[37];
    ks[11][14] = passwd[30];
    ks[11][15] = passwd[59];
    ks[11][16] = passwd[44];
    ks[11][17] = passwd[62];
    ks[11][18] = passwd[61];
    ks[11][19] = passwd[35];
    ks[11][20] = passwd[36];
    ks[11][21] = passwd[29];
    ks[11][22] = passwd[22];
    ks[11][23] = passwd[12];
    ks[11][24] = passwd[16];
    ks[11][25] = passwd[58];
    ks[11][26] = passwd[2];
    ks[11][27] = passwd[48];
    ks[11][28] = passwd[33];
    ks[11][29] = passwd[34];
    ks[11][30] = passwd[10];
    ks[11][31] = passwd[24];
    ks[11][32] = passwd[1];
    ks[11][33] = passwd[49];
    ks[11][34] = passwd[19];
    ks[11][35] = passwd[25];
    ks[11][36] = passwd[57];
    ks[11][37] = passwd[17];
    ks[11][38] = passwd[32];
    ks[11][39] = passwd[26];
    ks[11][40] = passwd[11];
    ks[11][41] = passwd[50];
    ks[11][42] = passwd[41];
    ks[11][43] = passwd[8];
    ks[11][44] = passwd[9];
    ks[11][45] = passwd[56];
    ks[11][46] = passwd[18];
    ks[11][47] = passwd[27];
    ks[12][0] = passwd[61];
    ks[12][1] = passwd[37];
    ks[12][2] = passwd[22];
    ks[12][3] = passwd[44];
    ks[12][4] = passwd[4];
    ks[12][5] = passwd[35];
    ks[12][6] = passwd[51];
    ks[12][7] = passwd[12];
    ks[12][8] = passwd[53];
    ks[12][9] = passwd[62];
    ks[12][10] = passwd[5];
    ks[12][11] = passwd[30];
    ks[12][12] = passwd[52];
    ks[12][13] = passwd[21];
    ks[12][14] = passwd[14];
    ks[12][15] = passwd[43];
    ks[12][16] = passwd[28];
    ks[12][17] = passwd[46];
    ks[12][18] = passwd[45];
    ks[12][19] = passwd[54];
    ks[12][20] = passwd[20];
    ks[12][21] = passwd[13];
    ks[12][22] = passwd[6];
    ks[12][23] = passwd[59];
    ks[12][24] = passwd[0];
    ks[12][25] = passwd[42];
    ks[12][26] = passwd[19];
    ks[12][27] = passwd[32];
    ks[12][28] = passwd[17];
    ks[12][29] = passwd[18];
    ks[12][30] = passwd[27];
    ks[12][31] = passwd[8];
    ks[12][32] = passwd[50];
    ks[12][33] = passwd[33];
    ks[12][34] = passwd[3];
    ks[12][35] = passwd[9];
    ks[12][36] = passwd[41];
    ks[12][37] = passwd[1];
    ks[12][38] = passwd[16];
    ks[12][39] = passwd[10];
    ks[12][40] = passwd[56];
    ks[12][41] = passwd[34];
    ks[12][42] = passwd[25];
    ks[12][43] = passwd[57];
    ks[12][44] = passwd[58];
    ks[12][45] = passwd[40];
    ks[12][46] = passwd[2];
    ks[12][47] = passwd[11];
    ks[13][0] = passwd[45];
    ks[13][1] = passwd[21];
    ks[13][2] = passwd[6];
    ks[13][3] = passwd[28];
    ks[13][4] = passwd[51];
    ks[13][5] = passwd[54];
    ks[13][6] = passwd[35];
    ks[13][7] = passwd[59];
    ks[13][8] = passwd[37];
    ks[13][9] = passwd[46];
    ks[13][10] = passwd[52];
    ks[13][11] = passwd[14];
    ks[13][12] = passwd[36];
    ks[13][13] = passwd[5];
    ks[13][14] = passwd[61];
    ks[13][15] = passwd[62];
    ks[13][16] = passwd[12];
    ks[13][17] = passwd[30];
    ks[13][18] = passwd[29];
    ks[13][19] = passwd[38];
    ks[13][20] = passwd[4];
    ks[13][21] = passwd[60];
    ks[13][22] = passwd[53];
    ks[13][23] = passwd[43];
    ks[13][24] = passwd[49];
    ks[13][25] = passwd[26];
    ks[13][26] = passwd[3];
    ks[13][27] = passwd[16];
    ks[13][28] = passwd[1];
    ks[13][29] = passwd[2];
    ks[13][30] = passwd[11];
    ks[13][31] = passwd[57];
    ks[13][32] = passwd[34];
    ks[13][33] = passwd[17];
    ks[13][34] = passwd[48];
    ks[13][35] = passwd[58];
    ks[13][36] = passwd[25];
    ks[13][37] = passwd[50];
    ks[13][38] = passwd[0];
    ks[13][39] = passwd[27];
    ks[13][40] = passwd[40];
    ks[13][41] = passwd[18];
    ks[13][42] = passwd[9];
    ks[13][43] = passwd[41];
    ks[13][44] = passwd[42];
    ks[13][45] = passwd[24];
    ks[13][46] = passwd[19];
    ks[13][47] = passwd[56];
    ks[14][0] = passwd[29];
    ks[14][1] = passwd[5];
    ks[14][2] = passwd[53];
    ks[14][3] = passwd[12];
    ks[14][4] = passwd[35];
    ks[14][5] = passwd[38];
    ks[14][6] = passwd[54];
    ks[14][7] = passwd[43];
    ks[14][8] = passwd[21];
    ks[14][9] = passwd[30];
    ks[14][10] = passwd[36];
    ks[14][11] = passwd[61];
    ks[14][12] = passwd[20];
    ks[14][13] = passwd[52];
    ks[14][14] = passwd[45];
    ks[14][15] = passwd[46];
    ks[14][16] = passwd[59];
    ks[14][17] = passwd[14];
    ks[14][18] = passwd[13];
    ks[14][19] = passwd[22];
    ks[14][20] = passwd[51];
    ks[14][21] = passwd[44];
    ks[14][22] = passwd[37];
    ks[14][23] = passwd[62];
    ks[14][24] = passwd[33];
    ks[14][25] = passwd[10];
    ks[14][26] = passwd[48];
    ks[14][27] = passwd[0];
    ks[14][28] = passwd[50];
    ks[14][29] = passwd[19];
    ks[14][30] = passwd[56];
    ks[14][31] = passwd[41];
    ks[14][32] = passwd[18];
    ks[14][33] = passwd[1];
    ks[14][34] = passwd[32];
    ks[14][35] = passwd[42];
    ks[14][36] = passwd[9];
    ks[14][37] = passwd[34];
    ks[14][38] = passwd[49];
    ks[14][39] = passwd[11];
    ks[14][40] = passwd[24];
    ks[14][41] = passwd[2];
    ks[14][42] = passwd[58];
    ks[14][43] = passwd[25];
    ks[14][44] = passwd[26];
    ks[14][45] = passwd[8];
    ks[14][46] = passwd[3];
    ks[14][47] = passwd[40];
    ks[15][0] = passwd[21];
    ks[15][1] = passwd[60];
    ks[15][2] = passwd[45];
    ks[15][3] = passwd[4];
    ks[15][4] = passwd[62];
    ks[15][5] = passwd[30];
    ks[15][6] = passwd[46];
    ks[15][7] = passwd[35];
    ks[15][8] = passwd[13];
    ks[15][9] = passwd[22];
    ks[15][10] = passwd[28];
    ks[15][11] = passwd[53];
    ks[15][12] = passwd[12];
    ks[15][13] = passwd[44];
    ks[15][14] = passwd[37];
    ks[15][15] = passwd[38];
    ks[15][16] = passwd[51];
    ks[15][17] = passwd[6];
    ks[15][18] = passwd[5];
    ks[15][19] = passwd[14];
    ks[15][20] = passwd[43];
    ks[15][21] = passwd[36];
    ks[15][22] = passwd[29];
    ks[15][23] = passwd[54];
    ks[15][24] = passwd[25];
    ks[15][25] = passwd[2];
    ks[15][26] = passwd[40];
    ks[15][27] = passwd[57];
    ks[15][28] = passwd[42];
    ks[15][29] = passwd[11];
    ks[15][30] = passwd[48];
    ks[15][31] = passwd[33];
    ks[15][32] = passwd[10];
    ks[15][33] = passwd[58];
    ks[15][34] = passwd[24];
    ks[15][35] = passwd[34];
    ks[15][36] = passwd[1];
    ks[15][37] = passwd[26];
    ks[15][38] = passwd[41];
    ks[15][39] = passwd[3];
    ks[15][40] = passwd[16];
    ks[15][41] = passwd[27];
    ks[15][42] = passwd[50];
    ks[15][43] = passwd[17];
    ks[15][44] = passwd[18];
    ks[15][45] = passwd[0];
    ks[15][46] = passwd[56];
    ks[15][47] = passwd[32];


    for (m = 0; m < 25; m ++) {

            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[0][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[0][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[0][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[0][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[0][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[0][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[0][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[0][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[0][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[0][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[0][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[0][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[0][12]),
               spu_xor(R1[8], ks[0][13]),
               spu_xor(R1[9], ks[0][14]),
               spu_xor(R1[10], ks[0][15]),
               spu_xor(R1[11], ks[0][16]),
               spu_xor(R1[12], ks[0][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[0][18]),
               spu_xor(R1[12], ks[0][19]),
               spu_xor(R1[13], ks[0][20]),
               spu_xor(R1[14], ks[0][21]),
               spu_xor(R1[15], ks[0][22]),
               spu_xor(R1[16], ks[0][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[0][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[0][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[0][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[0][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[0][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[0][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[0][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[0][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[0][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[0][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[0][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[0][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[0][36]),
               spu_xor(R1[24], ks[0][37]),
               spu_xor(R1[25], ks[0][38]),
               spu_xor(R1[26], ks[0][39]),
               spu_xor(R1[27], ks[0][40]),
               spu_xor(R1[28], ks[0][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[0][42]),
               spu_xor(R1[28], ks[0][43]),
               spu_xor(R1[29], ks[0][44]),
               spu_xor(R1[30], ks[0][45]),
               spu_xor(R1[31], ks[0][46]),
               spu_xor(R1[0], ks[0][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[1][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[1][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[1][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[1][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[1][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[1][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[1][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[1][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[1][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[1][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[1][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[1][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[1][12]),
               spu_xor(R0[8], ks[1][13]),
               spu_xor(R0[9], ks[1][14]),
               spu_xor(R0[10], ks[1][15]),
               spu_xor(R0[11], ks[1][16]),
               spu_xor(R0[12], ks[1][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[1][18]),
               spu_xor(R0[12], ks[1][19]),
               spu_xor(R0[13], ks[1][20]),
               spu_xor(R0[14], ks[1][21]),
               spu_xor(R0[15], ks[1][22]),
               spu_xor(R0[16], ks[1][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[1][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[1][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[1][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[1][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[1][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[1][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[1][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[1][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[1][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[1][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[1][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[1][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[1][36]),
               spu_xor(R0[24], ks[1][37]),
               spu_xor(R0[25], ks[1][38]),
               spu_xor(R0[26], ks[1][39]),
               spu_xor(R0[27], ks[1][40]),
               spu_xor(R0[28], ks[1][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[1][42]),
               spu_xor(R0[28], ks[1][43]),
               spu_xor(R0[29], ks[1][44]),
               spu_xor(R0[30], ks[1][45]),
               spu_xor(R0[31], ks[1][46]),
               spu_xor(R0[0], ks[1][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[2][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[2][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[2][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[2][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[2][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[2][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[2][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[2][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[2][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[2][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[2][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[2][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[2][12]),
               spu_xor(R1[8], ks[2][13]),
               spu_xor(R1[9], ks[2][14]),
               spu_xor(R1[10], ks[2][15]),
               spu_xor(R1[11], ks[2][16]),
               spu_xor(R1[12], ks[2][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[2][18]),
               spu_xor(R1[12], ks[2][19]),
               spu_xor(R1[13], ks[2][20]),
               spu_xor(R1[14], ks[2][21]),
               spu_xor(R1[15], ks[2][22]),
               spu_xor(R1[16], ks[2][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[2][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[2][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[2][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[2][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[2][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[2][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[2][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[2][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[2][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[2][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[2][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[2][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[2][36]),
               spu_xor(R1[24], ks[2][37]),
               spu_xor(R1[25], ks[2][38]),
               spu_xor(R1[26], ks[2][39]),
               spu_xor(R1[27], ks[2][40]),
               spu_xor(R1[28], ks[2][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[2][42]),
               spu_xor(R1[28], ks[2][43]),
               spu_xor(R1[29], ks[2][44]),
               spu_xor(R1[30], ks[2][45]),
               spu_xor(R1[31], ks[2][46]),
               spu_xor(R1[0], ks[2][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[3][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[3][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[3][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[3][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[3][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[3][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[3][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[3][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[3][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[3][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[3][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[3][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[3][12]),
               spu_xor(R0[8], ks[3][13]),
               spu_xor(R0[9], ks[3][14]),
               spu_xor(R0[10], ks[3][15]),
               spu_xor(R0[11], ks[3][16]),
               spu_xor(R0[12], ks[3][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[3][18]),
               spu_xor(R0[12], ks[3][19]),
               spu_xor(R0[13], ks[3][20]),
               spu_xor(R0[14], ks[3][21]),
               spu_xor(R0[15], ks[3][22]),
               spu_xor(R0[16], ks[3][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[3][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[3][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[3][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[3][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[3][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[3][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[3][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[3][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[3][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[3][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[3][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[3][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[3][36]),
               spu_xor(R0[24], ks[3][37]),
               spu_xor(R0[25], ks[3][38]),
               spu_xor(R0[26], ks[3][39]),
               spu_xor(R0[27], ks[3][40]),
               spu_xor(R0[28], ks[3][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[3][42]),
               spu_xor(R0[28], ks[3][43]),
               spu_xor(R0[29], ks[3][44]),
               spu_xor(R0[30], ks[3][45]),
               spu_xor(R0[31], ks[3][46]),
               spu_xor(R0[0], ks[3][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[4][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[4][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[4][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[4][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[4][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[4][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[4][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[4][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[4][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[4][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[4][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[4][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[4][12]),
               spu_xor(R1[8], ks[4][13]),
               spu_xor(R1[9], ks[4][14]),
               spu_xor(R1[10], ks[4][15]),
               spu_xor(R1[11], ks[4][16]),
               spu_xor(R1[12], ks[4][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[4][18]),
               spu_xor(R1[12], ks[4][19]),
               spu_xor(R1[13], ks[4][20]),
               spu_xor(R1[14], ks[4][21]),
               spu_xor(R1[15], ks[4][22]),
               spu_xor(R1[16], ks[4][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[4][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[4][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[4][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[4][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[4][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[4][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[4][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[4][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[4][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[4][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[4][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[4][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[4][36]),
               spu_xor(R1[24], ks[4][37]),
               spu_xor(R1[25], ks[4][38]),
               spu_xor(R1[26], ks[4][39]),
               spu_xor(R1[27], ks[4][40]),
               spu_xor(R1[28], ks[4][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[4][42]),
               spu_xor(R1[28], ks[4][43]),
               spu_xor(R1[29], ks[4][44]),
               spu_xor(R1[30], ks[4][45]),
               spu_xor(R1[31], ks[4][46]),
               spu_xor(R1[0], ks[4][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[5][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[5][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[5][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[5][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[5][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[5][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[5][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[5][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[5][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[5][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[5][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[5][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[5][12]),
               spu_xor(R0[8], ks[5][13]),
               spu_xor(R0[9], ks[5][14]),
               spu_xor(R0[10], ks[5][15]),
               spu_xor(R0[11], ks[5][16]),
               spu_xor(R0[12], ks[5][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[5][18]),
               spu_xor(R0[12], ks[5][19]),
               spu_xor(R0[13], ks[5][20]),
               spu_xor(R0[14], ks[5][21]),
               spu_xor(R0[15], ks[5][22]),
               spu_xor(R0[16], ks[5][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[5][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[5][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[5][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[5][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[5][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[5][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[5][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[5][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[5][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[5][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[5][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[5][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[5][36]),
               spu_xor(R0[24], ks[5][37]),
               spu_xor(R0[25], ks[5][38]),
               spu_xor(R0[26], ks[5][39]),
               spu_xor(R0[27], ks[5][40]),
               spu_xor(R0[28], ks[5][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[5][42]),
               spu_xor(R0[28], ks[5][43]),
               spu_xor(R0[29], ks[5][44]),
               spu_xor(R0[30], ks[5][45]),
               spu_xor(R0[31], ks[5][46]),
               spu_xor(R0[0], ks[5][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[6][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[6][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[6][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[6][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[6][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[6][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[6][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[6][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[6][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[6][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[6][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[6][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[6][12]),
               spu_xor(R1[8], ks[6][13]),
               spu_xor(R1[9], ks[6][14]),
               spu_xor(R1[10], ks[6][15]),
               spu_xor(R1[11], ks[6][16]),
               spu_xor(R1[12], ks[6][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[6][18]),
               spu_xor(R1[12], ks[6][19]),
               spu_xor(R1[13], ks[6][20]),
               spu_xor(R1[14], ks[6][21]),
               spu_xor(R1[15], ks[6][22]),
               spu_xor(R1[16], ks[6][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[6][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[6][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[6][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[6][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[6][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[6][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[6][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[6][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[6][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[6][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[6][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[6][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[6][36]),
               spu_xor(R1[24], ks[6][37]),
               spu_xor(R1[25], ks[6][38]),
               spu_xor(R1[26], ks[6][39]),
               spu_xor(R1[27], ks[6][40]),
               spu_xor(R1[28], ks[6][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[6][42]),
               spu_xor(R1[28], ks[6][43]),
               spu_xor(R1[29], ks[6][44]),
               spu_xor(R1[30], ks[6][45]),
               spu_xor(R1[31], ks[6][46]),
               spu_xor(R1[0], ks[6][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[7][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[7][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[7][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[7][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[7][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[7][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[7][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[7][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[7][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[7][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[7][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[7][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[7][12]),
               spu_xor(R0[8], ks[7][13]),
               spu_xor(R0[9], ks[7][14]),
               spu_xor(R0[10], ks[7][15]),
               spu_xor(R0[11], ks[7][16]),
               spu_xor(R0[12], ks[7][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[7][18]),
               spu_xor(R0[12], ks[7][19]),
               spu_xor(R0[13], ks[7][20]),
               spu_xor(R0[14], ks[7][21]),
               spu_xor(R0[15], ks[7][22]),
               spu_xor(R0[16], ks[7][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[7][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[7][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[7][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[7][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[7][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[7][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[7][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[7][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[7][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[7][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[7][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[7][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[7][36]),
               spu_xor(R0[24], ks[7][37]),
               spu_xor(R0[25], ks[7][38]),
               spu_xor(R0[26], ks[7][39]),
               spu_xor(R0[27], ks[7][40]),
               spu_xor(R0[28], ks[7][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[7][42]),
               spu_xor(R0[28], ks[7][43]),
               spu_xor(R0[29], ks[7][44]),
               spu_xor(R0[30], ks[7][45]),
               spu_xor(R0[31], ks[7][46]),
               spu_xor(R0[0], ks[7][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[8][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[8][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[8][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[8][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[8][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[8][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[8][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[8][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[8][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[8][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[8][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[8][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[8][12]),
               spu_xor(R1[8], ks[8][13]),
               spu_xor(R1[9], ks[8][14]),
               spu_xor(R1[10], ks[8][15]),
               spu_xor(R1[11], ks[8][16]),
               spu_xor(R1[12], ks[8][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[8][18]),
               spu_xor(R1[12], ks[8][19]),
               spu_xor(R1[13], ks[8][20]),
               spu_xor(R1[14], ks[8][21]),
               spu_xor(R1[15], ks[8][22]),
               spu_xor(R1[16], ks[8][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[8][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[8][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[8][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[8][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[8][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[8][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[8][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[8][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[8][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[8][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[8][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[8][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[8][36]),
               spu_xor(R1[24], ks[8][37]),
               spu_xor(R1[25], ks[8][38]),
               spu_xor(R1[26], ks[8][39]),
               spu_xor(R1[27], ks[8][40]),
               spu_xor(R1[28], ks[8][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[8][42]),
               spu_xor(R1[28], ks[8][43]),
               spu_xor(R1[29], ks[8][44]),
               spu_xor(R1[30], ks[8][45]),
               spu_xor(R1[31], ks[8][46]),
               spu_xor(R1[0], ks[8][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[9][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[9][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[9][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[9][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[9][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[9][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[9][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[9][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[9][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[9][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[9][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[9][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[9][12]),
               spu_xor(R0[8], ks[9][13]),
               spu_xor(R0[9], ks[9][14]),
               spu_xor(R0[10], ks[9][15]),
               spu_xor(R0[11], ks[9][16]),
               spu_xor(R0[12], ks[9][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[9][18]),
               spu_xor(R0[12], ks[9][19]),
               spu_xor(R0[13], ks[9][20]),
               spu_xor(R0[14], ks[9][21]),
               spu_xor(R0[15], ks[9][22]),
               spu_xor(R0[16], ks[9][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[9][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[9][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[9][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[9][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[9][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[9][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[9][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[9][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[9][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[9][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[9][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[9][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[9][36]),
               spu_xor(R0[24], ks[9][37]),
               spu_xor(R0[25], ks[9][38]),
               spu_xor(R0[26], ks[9][39]),
               spu_xor(R0[27], ks[9][40]),
               spu_xor(R0[28], ks[9][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[9][42]),
               spu_xor(R0[28], ks[9][43]),
               spu_xor(R0[29], ks[9][44]),
               spu_xor(R0[30], ks[9][45]),
               spu_xor(R0[31], ks[9][46]),
               spu_xor(R0[0], ks[9][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[10][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[10][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[10][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[10][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[10][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[10][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[10][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[10][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[10][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[10][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[10][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[10][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[10][12]),
               spu_xor(R1[8], ks[10][13]),
               spu_xor(R1[9], ks[10][14]),
               spu_xor(R1[10], ks[10][15]),
               spu_xor(R1[11], ks[10][16]),
               spu_xor(R1[12], ks[10][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[10][18]),
               spu_xor(R1[12], ks[10][19]),
               spu_xor(R1[13], ks[10][20]),
               spu_xor(R1[14], ks[10][21]),
               spu_xor(R1[15], ks[10][22]),
               spu_xor(R1[16], ks[10][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[10][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[10][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[10][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[10][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[10][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[10][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[10][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[10][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[10][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[10][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[10][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[10][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[10][36]),
               spu_xor(R1[24], ks[10][37]),
               spu_xor(R1[25], ks[10][38]),
               spu_xor(R1[26], ks[10][39]),
               spu_xor(R1[27], ks[10][40]),
               spu_xor(R1[28], ks[10][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[10][42]),
               spu_xor(R1[28], ks[10][43]),
               spu_xor(R1[29], ks[10][44]),
               spu_xor(R1[30], ks[10][45]),
               spu_xor(R1[31], ks[10][46]),
               spu_xor(R1[0], ks[10][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[11][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[11][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[11][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[11][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[11][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[11][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[11][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[11][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[11][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[11][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[11][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[11][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[11][12]),
               spu_xor(R0[8], ks[11][13]),
               spu_xor(R0[9], ks[11][14]),
               spu_xor(R0[10], ks[11][15]),
               spu_xor(R0[11], ks[11][16]),
               spu_xor(R0[12], ks[11][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[11][18]),
               spu_xor(R0[12], ks[11][19]),
               spu_xor(R0[13], ks[11][20]),
               spu_xor(R0[14], ks[11][21]),
               spu_xor(R0[15], ks[11][22]),
               spu_xor(R0[16], ks[11][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[11][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[11][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[11][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[11][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[11][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[11][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[11][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[11][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[11][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[11][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[11][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[11][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[11][36]),
               spu_xor(R0[24], ks[11][37]),
               spu_xor(R0[25], ks[11][38]),
               spu_xor(R0[26], ks[11][39]),
               spu_xor(R0[27], ks[11][40]),
               spu_xor(R0[28], ks[11][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[11][42]),
               spu_xor(R0[28], ks[11][43]),
               spu_xor(R0[29], ks[11][44]),
               spu_xor(R0[30], ks[11][45]),
               spu_xor(R0[31], ks[11][46]),
               spu_xor(R0[0], ks[11][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[12][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[12][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[12][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[12][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[12][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[12][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[12][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[12][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[12][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[12][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[12][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[12][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[12][12]),
               spu_xor(R1[8], ks[12][13]),
               spu_xor(R1[9], ks[12][14]),
               spu_xor(R1[10], ks[12][15]),
               spu_xor(R1[11], ks[12][16]),
               spu_xor(R1[12], ks[12][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[12][18]),
               spu_xor(R1[12], ks[12][19]),
               spu_xor(R1[13], ks[12][20]),
               spu_xor(R1[14], ks[12][21]),
               spu_xor(R1[15], ks[12][22]),
               spu_xor(R1[16], ks[12][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[12][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[12][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[12][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[12][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[12][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[12][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[12][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[12][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[12][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[12][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[12][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[12][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[12][36]),
               spu_xor(R1[24], ks[12][37]),
               spu_xor(R1[25], ks[12][38]),
               spu_xor(R1[26], ks[12][39]),
               spu_xor(R1[27], ks[12][40]),
               spu_xor(R1[28], ks[12][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[12][42]),
               spu_xor(R1[28], ks[12][43]),
               spu_xor(R1[29], ks[12][44]),
               spu_xor(R1[30], ks[12][45]),
               spu_xor(R1[31], ks[12][46]),
               spu_xor(R1[0], ks[12][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[13][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[13][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[13][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[13][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[13][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[13][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[13][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[13][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[13][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[13][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[13][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[13][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[13][12]),
               spu_xor(R0[8], ks[13][13]),
               spu_xor(R0[9], ks[13][14]),
               spu_xor(R0[10], ks[13][15]),
               spu_xor(R0[11], ks[13][16]),
               spu_xor(R0[12], ks[13][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[13][18]),
               spu_xor(R0[12], ks[13][19]),
               spu_xor(R0[13], ks[13][20]),
               spu_xor(R0[14], ks[13][21]),
               spu_xor(R0[15], ks[13][22]),
               spu_xor(R0[16], ks[13][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[13][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[13][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[13][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[13][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[13][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[13][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[13][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[13][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[13][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[13][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[13][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[13][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[13][36]),
               spu_xor(R0[24], ks[13][37]),
               spu_xor(R0[25], ks[13][38]),
               spu_xor(R0[26], ks[13][39]),
               spu_xor(R0[27], ks[13][40]),
               spu_xor(R0[28], ks[13][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[13][42]),
               spu_xor(R0[28], ks[13][43]),
               spu_xor(R0[29], ks[13][44]),
               spu_xor(R0[30], ks[13][45]),
               spu_xor(R0[31], ks[13][46]),
               spu_xor(R0[0], ks[13][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);


            s1(spu_xor(spu_sel(R1[31], R1[15], salt_flag[0]), ks[14][0]),
               spu_xor(spu_sel(R1[0], R1[16], salt_flag[1]), ks[14][1]),
               spu_xor(spu_sel(R1[1], R1[17], salt_flag[2]), ks[14][2]),
               spu_xor(spu_sel(R1[2], R1[18], salt_flag[3]), ks[14][3]),
               spu_xor(spu_sel(R1[3], R1[19], salt_flag[4]), ks[14][4]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[5]), ks[14][5]),
               &R0[8], &R0[16], &R0[22], &R0[30]);

            s2(spu_xor(spu_sel(R1[3], R1[19], salt_flag[6]), ks[14][6]),
               spu_xor(spu_sel(R1[4], R1[20], salt_flag[7]), ks[14][7]),
               spu_xor(spu_sel(R1[5], R1[21], salt_flag[8]), ks[14][8]),
               spu_xor(spu_sel(R1[6], R1[22], salt_flag[9]), ks[14][9]),
               spu_xor(spu_sel(R1[7], R1[23], salt_flag[10]), ks[14][10]),
               spu_xor(spu_sel(R1[8], R1[24], salt_flag[11]), ks[14][11]),
               &R0[12], &R0[27], &R0[1], &R0[17]);
        
            s3(spu_xor(R1[7], ks[14][12]),
               spu_xor(R1[8], ks[14][13]),
               spu_xor(R1[9], ks[14][14]),
               spu_xor(R1[10], ks[14][15]),
               spu_xor(R1[11], ks[14][16]),
               spu_xor(R1[12], ks[14][17]),
               &R0[23], &R0[15], &R0[29], &R0[5]);

            s4(spu_xor(R1[11], ks[14][18]),
               spu_xor(R1[12], ks[14][19]),
               spu_xor(R1[13], ks[14][20]),
               spu_xor(R1[14], ks[14][21]),
               spu_xor(R1[15], ks[14][22]),
               spu_xor(R1[16], ks[14][23]),
               &R0[25], &R0[19], &R0[9], &R0[0]);

            s5(spu_xor(spu_sel(R1[15], R1[31], salt_flag[0]), ks[14][24]),
               spu_xor(spu_sel(R1[16], R1[0], salt_flag[1]), ks[14][25]),
               spu_xor(spu_sel(R1[17], R1[1], salt_flag[2]), ks[14][26]),
               spu_xor(spu_sel(R1[18], R1[2], salt_flag[3]), ks[14][27]),
               spu_xor(spu_sel(R1[19], R1[3], salt_flag[4]), ks[14][28]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[5]), ks[14][29]),
               &R0[7], &R0[13], &R0[24], &R0[2]);

            s6(spu_xor(spu_sel(R1[19], R1[3], salt_flag[6]), ks[14][30]),
               spu_xor(spu_sel(R1[20], R1[4], salt_flag[7]), ks[14][31]),
               spu_xor(spu_sel(R1[21], R1[5], salt_flag[8]), ks[14][32]),
               spu_xor(spu_sel(R1[22], R1[6], salt_flag[9]), ks[14][33]),
               spu_xor(spu_sel(R1[23], R1[7], salt_flag[10]), ks[14][34]),
               spu_xor(spu_sel(R1[24], R1[8], salt_flag[11]), ks[14][35]),
               &R0[3], &R0[28], &R0[10], &R0[18]);
	    
            s7(spu_xor(R1[23], ks[14][36]),
               spu_xor(R1[24], ks[14][37]),
               spu_xor(R1[25], ks[14][38]),
               spu_xor(R1[26], ks[14][39]),
               spu_xor(R1[27], ks[14][40]),
               spu_xor(R1[28], ks[14][41]),
               &R0[31], &R0[11], &R0[21], &R0[6]);
	    
            s8(spu_xor(R1[27], ks[14][42]),
               spu_xor(R1[28], ks[14][43]),
               spu_xor(R1[29], ks[14][44]),
               spu_xor(R1[30], ks[14][45]),
               spu_xor(R1[31], ks[14][46]),
               spu_xor(R1[0], ks[14][47]),
               &R0[4], &R0[26], &R0[14], &R0[20]);


            s1(spu_xor(spu_sel(R0[31], R0[15], salt_flag[0]), ks[15][0]),
               spu_xor(spu_sel(R0[0], R0[16], salt_flag[1]), ks[15][1]),
               spu_xor(spu_sel(R0[1], R0[17], salt_flag[2]), ks[15][2]),
               spu_xor(spu_sel(R0[2], R0[18], salt_flag[3]), ks[15][3]),
               spu_xor(spu_sel(R0[3], R0[19], salt_flag[4]), ks[15][4]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[5]), ks[15][5]),
               &R1[8], &R1[16], &R1[22], &R1[30]);

            s2(spu_xor(spu_sel(R0[3], R0[19], salt_flag[6]), ks[15][6]),
               spu_xor(spu_sel(R0[4], R0[20], salt_flag[7]), ks[15][7]),
               spu_xor(spu_sel(R0[5], R0[21], salt_flag[8]), ks[15][8]),
               spu_xor(spu_sel(R0[6], R0[22], salt_flag[9]), ks[15][9]),
               spu_xor(spu_sel(R0[7], R0[23], salt_flag[10]), ks[15][10]),
               spu_xor(spu_sel(R0[8], R0[24], salt_flag[11]), ks[15][11]),
               &R1[12], &R1[27], &R1[1], &R1[17]);

            s3(spu_xor(R0[7], ks[15][12]),
               spu_xor(R0[8], ks[15][13]),
               spu_xor(R0[9], ks[15][14]),
               spu_xor(R0[10], ks[15][15]),
               spu_xor(R0[11], ks[15][16]),
               spu_xor(R0[12], ks[15][17]),
               &R1[23], &R1[15], &R1[29], &R1[5]);

            s4(spu_xor(R0[11], ks[15][18]),
               spu_xor(R0[12], ks[15][19]),
               spu_xor(R0[13], ks[15][20]),
               spu_xor(R0[14], ks[15][21]),
               spu_xor(R0[15], ks[15][22]),
               spu_xor(R0[16], ks[15][23]),
               &R1[25], &R1[19], &R1[9], &R1[0]);

            s5(spu_xor(spu_sel(R0[15], R0[31], salt_flag[0]), ks[15][24]),
               spu_xor(spu_sel(R0[16], R0[0], salt_flag[1]), ks[15][25]),
               spu_xor(spu_sel(R0[17], R0[1], salt_flag[2]), ks[15][26]),
               spu_xor(spu_sel(R0[18], R0[2], salt_flag[3]), ks[15][27]),
               spu_xor(spu_sel(R0[19], R0[3], salt_flag[4]), ks[15][28]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[5]), ks[15][29]),
               &R1[7], &R1[13], &R1[24], &R1[2]);

            s6(spu_xor(spu_sel(R0[19], R0[3], salt_flag[6]), ks[15][30]),
               spu_xor(spu_sel(R0[20], R0[4], salt_flag[7]), ks[15][31]),
               spu_xor(spu_sel(R0[21], R0[5], salt_flag[8]), ks[15][32]),
               spu_xor(spu_sel(R0[22], R0[6], salt_flag[9]), ks[15][33]),
               spu_xor(spu_sel(R0[23], R0[7], salt_flag[10]), ks[15][34]),
               spu_xor(spu_sel(R0[24], R0[8], salt_flag[11]), ks[15][35]),
               &R1[3], &R1[28], &R1[10], &R1[18]);
	    
            s7(spu_xor(R0[23], ks[15][36]),
               spu_xor(R0[24], ks[15][37]),
               spu_xor(R0[25], ks[15][38]),
               spu_xor(R0[26], ks[15][39]),
               spu_xor(R0[27], ks[15][40]),
               spu_xor(R0[28], ks[15][41]),
               &R1[31], &R1[11], &R1[21], &R1[6]);
	    
            s8(spu_xor(R0[27], ks[15][42]),
               spu_xor(R0[28], ks[15][43]),
               spu_xor(R0[29], ks[15][44]),
               spu_xor(R0[30], ks[15][45]),
               spu_xor(R0[31], ks[15][46]),
               spu_xor(R0[0], ks[15][47]),
               &R1[4], &R1[26], &R1[14], &R1[20]);

        vptmp = R1;
        R1 = R0;
        R0 = vptmp;
    }

#define STORE_R() {                             \
        block[6] = R1[31];                      \
        block[7] = R0[31];                      \
        block[8] = R1[6];                       \
        block[9] = R0[6];                       \
        block[10] = R1[14];                     \
        block[11] = R0[14];                     \
        block[12] = R1[22];                     \
        block[13] = R0[22];                     \
        block[14] = R1[30];                     \
        block[15] = R0[30];                     \
        block[16] = R1[5];                      \
        block[17] = R0[5];                      \
        block[18] = R1[13];                     \
        block[19] = R0[13];                     \
        block[20] = R1[21];                     \
        block[21] = R0[21];                     \
        block[22] = R1[29];                     \
        block[23] = R0[29];                     \
        block[24] = R1[4];                      \
        block[25] = R0[4];                      \
        block[26] = R1[12];                     \
        block[27] = R0[12];                     \
        block[28] = R1[20];                     \
        block[29] = R0[20];                     \
        block[30] = R1[28];                     \
        block[31] = R0[28];                     \
                                                \
        block[32] = R1[3];                      \
        block[33] = R0[3];                      \
        block[34] = R1[11];                     \
        block[35] = R0[11];                     \
        block[36] = R1[19];                     \
        block[37] = R0[19];                     \
        block[38] = R1[27];                     \
        block[39] = R0[27];                     \
        block[40] = R1[2];                      \
        block[41] = R0[2];                      \
        block[42] = R1[10];                     \
        block[43] = R0[10];                     \
        block[44] = R1[18];                     \
        block[45] = R0[18];                     \
        block[46] = R1[26];                     \
        block[47] = R0[26];                     \
        block[48] = R1[1];                      \
        block[49] = R0[1];                      \
        block[50] = R1[9];                      \
        block[51] = R0[9];                      \
        block[52] = R1[17];                     \
        block[53] = R0[17];                     \
        block[54] = R1[25];                     \
        block[55] = R0[25];                     \
        block[56] = R1[0];                      \
        block[57] = R0[0];                      \
        block[58] = R1[8];                      \
        block[59] = R0[8];                      \
        block[60] = R1[16];                     \
        block[61] = R0[16];                     \
        block[62] = R1[24];                     \
        block[63] = R0[24];                     \
                                                \
    }

    STORE_R();

    
    /* ʸʸ̤ѷ */
    if (unlikely(ignore_case)) {

        vec_uchar16 sflag;
        vec_uchar16 carry1, carry2;

        for (i = 1; i < 11; i ++) {
            k = i * 6;
            /* ޤa(0x26)ʾӽФ */
            sflag = spu_and(block[k],
                            spu_or(spu_or(block[k + 1], block[k + 2]),
                                   spu_and(block[k + 3], block[k + 4])));

            /* sflag bitäƤˡ0x26ä */
            /* block[5] ϤΤޤ */

            /* block[4] 1­ */
            carry1 = spu_and(block[k + 4], sflag);

            /* 0sflag=1ʤ11ΤȤ0ˤʤ */
            block[k + 4] = spu_xor(block[k + 4], sflag); 

            /* block[3] 1­ */
            /* Ȥ1ä꤫꡼ϡɬ夬 */
            carry2 = spu_and(spu_or(block[k + 3], carry1), sflag);

            /* 1ǥ꡼10ǥ꡼0ΤȤ1ˤʤ  */
            block[k + 3] = spu_sel(block[k + 3], 
                                   spu_xor(spu_xor(block[k + 3], carry1), VFF),
                                   sflag);

            /* block[2] Ϸ夬Τ */
            /* 1ǡ꡼1ä鷫夬 */
            carry1 = spu_and(block[k + 2], carry2);
            block[k + 2] = spu_xor(block[k + 2], carry2);

            /* block[1] ⷫ夬Τ */
            carry2 = spu_and(block[k + 1], carry1);
            block[k + 1] = spu_xor(block[k + 1], carry1);

            /* block[0] */
            /* sflag xor carry xor block  */
            block[k] = spu_xor(spu_xor(block[k], carry2), sflag);
        }
    }



    /*
      ȥåפ10ʸʤΤǡ10̤ΰ֤Ф褤
     */
    vec_uchar16 match_tmp[10];
    vec_uchar16 match = VZERO;
    unsigned int matched;

    /* Ф˥ޥåʤʬϺǽ餫0 */
    for (i = 0; i < 10; i ++)
        match_tmp[i] = (i < 11 - pattern_len) ? VFF : VZERO;

    /* pattern_len110 (0ǥƥȥ⡼) */
#if 0
    for (i = 0; i < pattern_len * 6; i ++)
        for (j = 0; j < 11 - pattern_len; j ++)
            match_tmp[j] = spu_and(match_tmp[j], 
                                   spu_eqv(block[i + 6 * (j + 1)], pattern[i]));
#else
    for (i = 0; i < pattern_len * 6; i ++)
        for (j = 0; j < 10; j ++)
            match_tmp[j] = spu_and(match_tmp[j], 
                                   spu_eqv(block[i + 6 * (j + 1)], pattern[i]));
#endif

    for (i = 0; i < 10; i ++)
        match = spu_or(match, match_tmp[i]);

    match = spu_and(valid_range, match);

    matched = spu_extract(spu_orx(((vec_uint4)match)), 0);


    if (matched | (pattern_len == 0)) {

        int bit_offset = 0;

        match = pattern_len ? match : (vec_uchar16)valid_range;

        /* ʸʸ̤ϺR򥳥ԡ */
        if (unlikely(ignore_case))
            STORE_R();
        
        for (m = 0; m < 16; m ++) {
            unsigned int n = (unsigned int)spu_extract(match, m);
            int tail;
            for (k = 0; k < 8; k ++) {
                if ((n >> (7 - k)) & 1) {

                    mfc_write_tag_mask(DMA_TAG_RESULT_MASK);
                    mfc_read_tag_status_all();

                    for (i = 1; i < 11; i ++) {

                        result.trip[i - 1] = 0;

                        for(j = 0; j < 6; j ++) {
                            result.trip[i - 1] <<= 1;
                            result.trip[i - 1] 
                                |= (spu_extract(block[6 * i + j], m)
                                    >> (7 - k)) & 1;
                        }

                        result.trip[i - 1] += 46;
                        if (result.trip[i - 1] > 57)
                            result.trip[i - 1] += 7;
                        if (result.trip[i - 1] > 90)
                            result.trip[i - 1] += 6;

                    }

                    for (i = 0; i < 6; i ++)
                        result.password[i] = password[i];
                    
                    result.password[6] = ((offset + bit_offset) / 96) + ' ';
                    result.password[7] = ((offset + bit_offset) % 96) + ' ';
                    /* 96ܤκǸʸNULL */
                    result.password[7] = (result.password[7] > '~') 
                        ? '\0' : result.password[7];
                    
                    result.salt[0] = (unsigned char)salt[0];
                    result.salt[1] = (unsigned char)salt[1];
                    

                    tail = tail_pointer.p;

                    /* 1­ */
                    tail_pointer.p ++;
                    tail_pointer.p = (tail_pointer.p >= RESULT_BUF_LEN) 
                        ? 0 : tail_pointer.p;

                    /* ­ɤĤƤޤä顤ǿͤɤ߽Ф */
                    while (unlikely(head_pointer.p == tail_pointer.p)) {
                        spu_mfcdma32(&head_pointer, head_pointer_ea, 
                                     sizeof(result_pointer_t), DMA_TAG_BLOCK, 
                                     MFC_GET_CMD);
                        mfc_write_tag_mask(DMA_TAG_BLOCK_MASK);
                        mfc_read_tag_status_all();
                    }


#ifdef SPE_DEBUG
#if 0
                    printf("SPE[%d]: trip:%s password:%s\n", id, result.trip,
                           result.password);
#endif
#endif

                    spu_mfcdma32(&result, 
                                 result_buf_ea + sizeof(result_t) * tail, 
                                 sizeof(result_t), 
                                 DMA_TAG_RESULT, MFC_PUT_CMD);
                    

                    /* ɬˤɬפΤfence */
                    spu_mfcdma32(&tail_pointer, tail_pointer_ea, 
                                 sizeof(result_pointer_t), DMA_TAG_RESULT,
                                 MFC_PUTF_CMD);

                }
                bit_offset ++;
            }
        }
    }
}


/*
  saltʸ
*/
static inline int
correct_salt(int c)
{
    int out;
    
    out = (c < '.')
        | ((c > '9') & (c < 'A'))
        | ((c > 'Z') & (c < 'a'))
        | (c > 'z');

    return out ? '.' : c;
}

static inline int
salt2num(int c)
{
    c = c > 'Z' ? (c - 6) : c;
    c = c > '9' ? (c - 7) : c;

    c -= '.';

    return c;
}

/*
 * ѥɤƬ6ʸꤷơcoreƤӽФ
 *   (coreǡ2ʸѥ)
 */

static inline void
trip_search(const unsigned char *password, const vec_uchar16 *pattern, 
            const int pattern_len, const int ignore_case, const int kana_mode,
            const host_ea ea)
{
    int i, j;
    unsigned char c;
    unsigned int n = 0;

    unsigned int proc_num_ea = ea.proc_num_ea;

    static vec_uchar16 salt_flag[12];
    static int salt_cache[2] = {-1, -1};
    static int salt[2];


    /* ѥɤ64bit */
    vec_uchar16 passwd[64];

    int offset = 0;
    int p[2];

    p[0] = kana_mode ? (int)'.' : (int)password[1];
    p[1] = kana_mode ? (int)'.' : (int)password[2];

    /* salt */
    if (unlikely((p[0] != salt_cache[0]) | ( p[1] != salt_cache[1]))) {
        for (i = 0; i < 2; i ++) {
            int salti;
            salt_cache[i] = p[i];

            salt[i] = correct_salt(p[i]);

            salti = salt2num(salt[i]);

            for (j = 0; j < 6; j ++)
                salt_flag[i * 6 + j] = ((salti >> j) & 1) ? VFF : VZERO;
        }
    }

    for (i = 0; i < 6; i ++) {
        c = password[i];
        for (j = 0; j < 8; j ++) {
            passwd[i * 8 + j] = ((c >> j) & 1) ? VFF : VZERO;
        }
    }

    passwd[55] = VZERO;
    passwd[63] = VZERO;

    
    for (n = 0; n < 23; n ++) {

        /* 2ʸܤξ6bit沼5bitʬ */
        passwd[50] = (n & 1) ? VFF : VZERO;
        passwd[51] = (n & 2) ? VFF : VZERO;
        passwd[52] = (n & 4) ? VFF : VZERO;
        passwd[53] = (n & 8) ? VZERO : VFF;
        passwd[54] = (n >= 8) ? VFF : VZERO;


        /* 2ʸܤβ2bitʬ */
        passwd[48] = (vec_uchar16)((vec_uint4){0, 0, 0, ~0U});
        passwd[49] = VZERO;

        passwd[56] = PWD600;
        passwd[57] = PWD601;
        passwd[58] = PWD602;
        passwd[59] = PWD603;
        passwd[60] = PWD604;
        passwd[61] = PWD605;
        passwd[62] = PWD606;

        trip_search_core(passwd, salt_flag, pattern, pattern_len, ignore_case,
                         VFF, password, offset, salt, ea);

        offset += 128;

        passwd[48] = (vec_uchar16)((vec_uint4){~0U, ~0U, 0, 0});
        passwd[49] = (vec_uchar16)((vec_uint4){0, 0, ~0U, ~0U});

        passwd[56] = PWD610;
        passwd[57] = PWD611;
        passwd[58] = PWD612;
        passwd[59] = PWD613;
        passwd[60] = PWD614;
        passwd[61] = PWD615;
        passwd[62] = PWD616;

        trip_search_core(passwd, salt_flag, pattern, pattern_len, ignore_case,
                         VFF, password, offset, salt, ea);

        offset += 128;

        passwd[48] = (vec_uchar16)((vec_uint4){0, ~0U, ~0U, ~0U});
        passwd[49] = VFF;

        passwd[56] = PWD620;
        passwd[57] = PWD621;
        passwd[58] = PWD622;
        passwd[59] = PWD623;
        passwd[60] = PWD624;
        passwd[61] = PWD625;
        passwd[62] = PWD626;

        trip_search_core(passwd, salt_flag, pattern, pattern_len, ignore_case, 
                         VFF, password, offset, salt, ea);


        offset += 128;
    }

    /* n = 23ΤȤϺǸΰΤƤ (7fˤʤä㤦Τ) */
    n = 23;

    /* 2ʸܤξ6bit沼5bitʬ */
    passwd[50] = (n & 1) ? VFF : VZERO;
    passwd[51] = (n & 2) ? VFF : VZERO;
    passwd[52] = (n & 4) ? VFF : VZERO;
    passwd[53] = (n & 8) ? VZERO : VFF;
    passwd[54] = (n >= 8) ? VFF : VZERO;


    /* 2ʸܤβ2bitʬ */
    passwd[48] = (vec_uchar16)((vec_uint4){0, 0, 0, ~0U});
    passwd[49] = VZERO;

    passwd[56] = PWD600;
    passwd[57] = PWD601;
    passwd[58] = PWD602;
    passwd[59] = PWD603;
    passwd[60] = PWD604;
    passwd[61] = PWD605;
    passwd[62] = PWD606;


    trip_search_core(passwd, salt_flag, pattern, pattern_len, ignore_case,
                     VFF, password, offset, salt, ea);

    offset += 128;

    passwd[48] = (vec_uchar16)((vec_uint4){~0U, ~0U, 0, 0});
    passwd[49] = (vec_uchar16)((vec_uint4){0, 0, ~0U, ~0U});

    passwd[56] = PWD610;
    passwd[57] = PWD611;
    passwd[58] = PWD612;
    passwd[59] = PWD613;
    passwd[60] = PWD614;
    passwd[61] = PWD615;
    passwd[62] = PWD616;


    trip_search_core(passwd, salt_flag, pattern, pattern_len, ignore_case,
                     VFF, password, offset, salt, ea);


    offset += 128;

    passwd[48] = (vec_uchar16)((vec_uint4){0, ~0U, ~0U, ~0U});
    passwd[49] = VFF;

    passwd[56] = PWD620;
    passwd[57] = PWD621;
    passwd[58] = PWD622;
    passwd[59] = PWD623;
    passwd[60] = PWD624;
    passwd[61] = PWD625;
    passwd[62] = PWD626;


    trip_search_core(passwd, salt_flag, pattern, pattern_len, ignore_case, 
                     (vec_uchar16)((vec_uint4){~0U, 0, 0, 0}),
                     password, offset, salt, ea);


    mfc_write_tag_mask(DMA_TAG_PNUM);
    mfc_read_tag_status_all();

    proc_num_buf[0] += offset + 32;

    spu_mfcdma32(&proc_num_buf, 
                 proc_num_ea, 
                 sizeof(proc_num_buf), 
                 DMA_TAG_PNUM, MFC_PUT_CMD);
    
}


static inline void
trip_search_block(unsigned char *password, const vec_uchar16 *pattern, 
                  const int pattern_len, const int ignore_case, 
                  const int kana_mode, int distance, const host_ea ea)
{
    /* ѥɤβ3ʸ֤ѤѥĴ */
    while (1) {

        trip_search(password, pattern, pattern_len, ignore_case,
                    kana_mode, ea);
        
        password[5] += distance;

        if (unlikely(password[5] > '~'))
            return;
    }
}


int
main(unsigned long long argc, unsigned long long argv)
{
    unsigned char password[16] = {0};


    unsigned int cmd;
    unsigned int ea;
    int loop = 1;
    unsigned int control_flag_ea = 0;
    int ignore_case = 0;

    int pattern_len = 0;

    int kana_mode = 0;
    int i;

    int distance = 1;

    long long int block_num = -1;

    host_ea h_ea = {0};

    while (loop) {
        while (spu_stat_in_mbox() == 0)
            ;

        cmd = spu_read_in_mbox();

        switch (cmd) {

        case CMD_SET_ID:
            while (spu_stat_in_mbox() == 0)
                ;
            id = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("ID %d\n", id);
#endif

            break;

        case CMD_LOAD_PATTERN: 
            while (spu_stat_in_mbox() == 0)
                ;
            ea = spu_read_in_mbox();

            spu_mfcdma32(pattern,
                         ea, sizeof(vec_uchar16) * 64, 
                         DMA_TAG_BLOCK, MFC_GET_CMD);
            mfc_write_tag_mask(DMA_TAG_BLOCK_MASK);
            mfc_read_tag_status_all();


            /*
            for (i = 0; i < 64; i ++) {
                printf("%2d", i);
                dump_vector("", pattern[i], 64);
            }
            */

            break;

        case CMD_SET_PATTERN_LEN:
            while (spu_stat_in_mbox() == 0)
                ;
            pattern_len = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("SPE[%d] pattern_len %d\n", id, pattern_len);
#endif

            break;

        case CMD_LOAD_BLOCK_NUM: 
            while (spu_stat_in_mbox() == 0)
                ;
            ea = spu_read_in_mbox();
            
            spu_mfcdma32(block_num_buf,
                         ea, sizeof(long long int) * 2, 
                         DMA_TAG_BLOCK, MFC_GET_CMD);
            mfc_write_tag_mask(DMA_TAG_BLOCK_MASK);
            mfc_read_tag_status_all();

            block_num = block_num_buf[0];

#ifdef SPE_DEBUG
            printf("SPE[%d] block_num: %lld\n", id, block_num);
#endif

            break;

        case CMD_SET_CASE_FLAG:
            ignore_case = 1;

#ifdef SPE_DEBUG
            printf("SPE[%d] ignore case %d\n", id, ignore_case);
#endif

            break;

        case CMD_SET_CONTROL_EA:
            while (spu_stat_in_mbox() == 0)
                ;
            control_flag_ea = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("SPE[%d] control flag ea 0x%08x\n", id, control_flag_ea);
#endif

            break;

        case CMD_SET_PROC_NUM_EA:
            while (spu_stat_in_mbox() == 0)
                ;
            h_ea.proc_num_ea = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("SPE[%d] processed_num_buf EA 0x%08x\n", id, 
                   h_ea.proc_num_ea);
#endif

            break;


        case CMD_SET_RESULT_EA:
            while (spu_stat_in_mbox() == 0)
                ;
            h_ea.result_buf_ea = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("SPE[%d] result ea 0x%08x\n", id, h_ea.result_buf_ea);
#endif

            break;

        case CMD_SET_HEAD_PTR_EA:
            while (spu_stat_in_mbox() == 0)
                ;
            h_ea.head_pointer_ea = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("SPE[%d] head ea 0x%08x\n", id, h_ea.head_pointer_ea);
#endif

            break;


        case CMD_SET_TAIL_PTR_EA:
            while (spu_stat_in_mbox() == 0)
                ;
            h_ea.tail_pointer_ea = spu_read_in_mbox();

#ifdef SPE_DEBUG
            printf("SPE[%d] tail ea 0x%08x\n", id, h_ea.tail_pointer_ea);
#endif

            break;


        case CMD_SET_DISTANCE:
            while (spu_stat_in_mbox() == 0)
                ;
            distance = spu_read_in_mbox();

            if (distance <= 0 || distance > ('~' - ' '))
                distance = 1;

#ifdef SPE_DEBUG
            printf("SPE[%d] distance %d\n", id, distance);
#endif

            break;


        case CMD_LOAD_INIT_PASS:
            while (spu_stat_in_mbox() == 0)
                ;
            ea = spu_read_in_mbox();
            spu_mfcdma32(ipassword, ea, sizeof(ipassword), 
                         DMA_TAG_BLOCK,  MFC_GET_CMD);
            mfc_write_tag_mask(DMA_TAG_BLOCK_MASK);
            mfc_read_tag_status_all();

            for (i = 0; i < 8; i ++)
                password[i] = ipassword[i];

#ifdef SPE_DEBUG
            printf("SPE[%d] initial password: %s\n", id, ipassword);
#endif

            break;


        case CMD_SET_KANA_MODE:
            kana_mode = 1;

#ifdef SPE_DEBUG
            printf("SPE[%d] kana mode enabled\n", id);
#endif

            break;


        case CMD_RUN: 

            while (1) {

                /* main part */
                spu_mfcdma32(control_flag, control_flag_ea, 
                             sizeof(control_flag), DMA_TAG_NONBLOCK,
                             MFC_GET_CMD);

                if (unlikely((block_num == 0)))
                    break;

                trip_search_block(password, pattern, pattern_len, ignore_case,
                                  kana_mode, distance, h_ea);

                mfc_write_tag_mask(DMA_TAG_NONBLOCK_MASK);
                mfc_read_tag_status_all();

                block_num = (block_num > 0) ? (block_num - 1) : block_num;


                if (unlikely((control_flag[0]))) {
                    /* ե饰äƤȴ */
                    if (control_flag[0] == CONTROL_STOP) {
#ifdef SPE_DEBUG
                        printf("SPE[%d] catch\n", id);
#endif
                        break;
                    } 
                }

                password[5] = ' ' + (password[5] - '~' - 1);
                password[4] ++;
                if (unlikely((password[4] > '~'))) {
                    password[4] = ' ';
                    password[3] ++;
                    if (unlikely((password[3] > '~'))) {
                        password[3] = ' ';
                        password[2] ++;
                        if (unlikely((password[2] > '~'))) {
                            password[2] = ' ';
                            password[1] ++;
                            if (unlikely((password[1] > '~'))) {
                                password[1] = ' ';
                                password[0] ++;
                                if (unlikely((password[0] > '~'))) {
                                    password[0] = ' ';
                                }
                            }
                        }   
                    }
                }
            }
            
            break;



        case CMD_EXIT:
            loop = 0;
            break;
        }
        
    }


    /* ǸǤФʬäƤ뤳ȤǧƤ */
    mfc_write_tag_mask(DMA_TAG_ALL_MASK);
    mfc_read_tag_status_all();

#ifdef SPE_DEBUG
    printf("SPE[%d] done\n", id);
#endif

    return 0;
}


